from dataclasses import dataclass
from typing import Optional, Union, Literal

from utz import err, check, process

from ire.utils import run


def git_has_staged_changes():
    return not check('git', 'diff', '--cached', '--quiet', 'HEAD', log=None)


def git_root():
    return process.line('git', 'rev-parse', '--show-toplevel', log=None)


def current_sha():
    return process.line('git', 'rev-parse', '--short', 'HEAD', log=None)


def tracking_sha():
    return process.line('git', 'rev-parse', '--short', '@{u}', log=None)


def has_unpushed_commits():
    return current_sha() != tracking_sha()


@dataclass
class CommitConfig:
    allow_existing_staged_changes: bool = False
    verbose: bool = False

    def __call__(self, path):
        self.commit(path)

    def commit(self, path):
        if self.allow_existing_staged_changes or not git_has_staged_changes():
            run('git', 'add', '-f', path, verbose=self.verbose)
            if git_has_staged_changes():
                run('git', 'commit', '-m', f'iRe: add {path}', verbose=self.verbose)
            else:
                err(f"iRe: {path} appears unchanged, `git add` had no effect")
        else:
            raise RuntimeError(
                "Refusing to Git commit given existing staged changes (passing `git=commit!` will add the new "
                "table to staged changes, and commit)"
            )


@dataclass
class PushConfig:
    commit: Optional[CommitConfig] = None
    remote: Optional[str] = None
    verbose: bool = False

    def push(self):
        remote = self.remote
        run(['git', 'push'] + [ remote ] if remote else [], verbose=self.verbose)


GitConfig = Union[
    Literal['add', 'commit', 'commit!', 'push'],
    CommitConfig,
    PushConfig,
]


def process_git_config(git: Optional[GitConfig], path: str, verbose: bool = False):
    """Process the git config: add, commit, or push the file `path`."""
    git_configs = {
        'commit': CommitConfig(verbose=verbose),
        'commit!': CommitConfig(allow_existing_staged_changes=True, verbose=verbose),
        'push': PushConfig(commit=CommitConfig(verbose=verbose), verbose=verbose),
        'push!': PushConfig(commit=CommitConfig(allow_existing_staged_changes=True, verbose=verbose), verbose=verbose),
    }
    if isinstance(git, str) and git in git_configs:
        git = git_configs[git]
    if git == 'add':
        run('git', 'add', '-f', path, verbose=verbose)
    elif isinstance(git, CommitConfig):
        git.commit(path)
    elif isinstance(git, PushConfig):
        git.commit(path)
        run('git', 'push', verbose=verbose)



import hashlib
import shlex
from subprocess import Popen, PIPE
import sys

from IPython import extract_module_locals


def md5_file(fname):
    """Compute the MD5 hash of a file's contents

    https://stackoverflow.com/a/3431838
    """
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def cur_cell_code(depth=0):
    lcls = extract_module_locals(depth=depth)[1]
    keys = list(lcls.keys())
    print(f"{keys=}")
    return lcls[keys[-1]]


def run(*cmd, verbose: bool = False):
    """Run a command, capturing stdout/stderr, printing iff `verbose` or a failure occurs."""
    proc = Popen(cmd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = proc.communicate()
    cmd_str = shlex.join(cmd)

    def print_streams():
        nonlocal stdout, stderr
        if stdout:
            # For some reason `dvc push` stdout sometimes comes back as b'Everything is up to date.\x1b[0m'.
            # TODO: debug (possibly login shell interference?)
            ansi_suffix = b'\x1b[0m'
            if stdout.endswith(ansi_suffix):
                stdout = stdout[:-len(ansi_suffix)]
            try:
                stdout = stdout.decode()
            except UnicodeDecodeError:
                pass
            sys.stdout.write(f'{cmd_str}: {stdout}')
        if stderr:
            try:
                stderr = stderr.decode()
            except UnicodeDecodeError:
                pass
            sys.stderr.write(f'{cmd_str}: {stderr}')

    if proc.returncode != 0:
        print_streams()
        raise RuntimeError(f"{cmd_str} failed ({proc.returncode}): {stderr}")
    elif verbose:
        print_streams()



from os.path import join, exists
from subprocess import CalledProcessError
from typing import Optional, Union

from .git import git_root
from .utils import run


def is_dvc_repo():
    try:
        git = git_root()
    except CalledProcessError:
        return False
    dvc_dir = join(git, '.dvc')
    return exists(dvc_dir)


# True ⟹ "add"
DvcConfig = Union[bool, 'add', 'push']


def process_dvc_config(dvc: Optional[DvcConfig], path: str, verbose: bool = False):
    """Process the dvc config, returning the path to the dvc file if it was created."""
    def run_dvc(cmd):
        run('dvc', cmd, path, verbose=verbose)

    if dvc:
        run_dvc('add')
        dvc_path = f'{path}.dvc'
        if dvc == 'push':
            run_dvc('push')
        return dvc_path
    return None



from setuptools import setup, find_packages

setup(
    name="iRe.py",
    version="0.1.1",
    description="Export tables and plots from Jupyter notebooks, along with metadata for embedding interactive tables in downstream apps.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=open("requirements.txt", "r").read(),
    extras_require={
        "plotly": "plotly",
        "test": open("requirements-dev.txt", "r").read(),
    },
    license="MIT",
    author="Ryan Williams",
    author_email="ryan@runsascoded.com",
    url="https://gitlab.com/runsascoded/ire/py",
)



import os
from dataclasses import dataclass
from inspect import getfullargspec
from os import getcwd, makedirs, rename
from os.path import islink, join
from tempfile import TemporaryDirectory
from typing import Literal, Optional, Union, TypeVar, Callable

import pandas as pd
from IPython.display import display, HTML, Image
from utz import err

from ire.dvc import DvcConfig, is_dvc_repo, process_dvc_config
from ire.git import GitConfig, has_unpushed_commits, process_git_config
from ire.utils import md5_file, run


DEFAULT_ARTIFACTS_DIR = ".ire"
METADATA_KEY = 'iRe'

ParquetEngine = Optional[Literal['fastparquet', 'pyarrow', 'auto']]


@dataclass
class Config:
    git: GitConfig
    dvc: DvcConfig
    dir: str = DEFAULT_ARTIFACTS_DIR
    engine: ParquetEngine = None


# Default config values
config = Config(
    git='add',
    dvc='add' if is_dvc_repo() else False,
)


@dataclass
class Output:
    path: str
    md5: str


Fmt = Union[Literal['parquet'], Literal['plotly'], Literal['png']]
ShowPlot = Union[None, 'png', 'html']


T = TypeVar("T")
DEFAULT_EXTENSION_OVERRIDES = {
    'plotly': 'json',
}


def chdir(path):
    os.chdir(path)
    config.dir = join(path, DEFAULT_ARTIFACTS_DIR)


def write_obj(
        obj: T,
        write: Callable,
        fmt: Fmt,
        name: str = None,
        dir: str = DEFAULT_ARTIFACTS_DIR,
        extension: str = None,
        git: Optional[GitConfig] = None,
        dvc: Optional[DvcConfig] = None,
        verbose: bool = False,
        engine: ParquetEngine = None,
):
    if extension is None:
        extension = DEFAULT_EXTENSION_OVERRIDES.get(fmt, fmt)

    def pre_write(path: str):
        if dvc and islink(path):
            # If the file is a symlink, we need to unprotect it before writing
            run('dvc', 'unprotect', path, verbose=verbose)

    write_kwargs = {}
    spec = getfullargspec(write)
    if 'engine' in spec.args:
        if engine is None:
            engine = config.engine
        if engine:
            write_kwargs['engine'] = engine

    if name is None:
        with TemporaryDirectory(dir=getcwd()) as tmpdir:
            basename = f"tmp.{extension}"
            tmp_path = os.path.join(tmpdir, basename)
            write(obj, tmp_path, **write_kwargs)
            md5 = md5_file(tmp_path)
            path = os.path.join(dir, f"{md5}.{extension}")
            makedirs(dir, exist_ok=True)
            pre_write(path)
            rename(tmp_path, path)
    else:
        path = os.path.join(dir, f"{name}.{extension}")
        makedirs(dir, exist_ok=True)
        pre_write(path)
        write(obj, path, **write_kwargs)
    err(f"Saved {fmt} to {path}")
    path = process_dvc_config(dvc, path, verbose=verbose) or path
    process_git_config(git, path, verbose=verbose)
    metadata = { METADATA_KEY: { 'fmt': fmt, 'path': path, 'dvc': bool(dvc), } }
    return metadata


def write_parquet(df: pd.DataFrame, path: str, engine: ParquetEngine = None):
    if engine is None:
        engine = 'auto'
    elif engine not in ('fastparquet', 'pyarrow'):
        raise ValueError(f"Unrecognized Parquet engine: {engine}")
    df.to_parquet(path, engine=engine)


def write_image(img: Union[Image, bytes], path: str):
    with open(path, 'wb') as f:
        data = img.data if isinstance(img, Image) else img
        f.write(data)


def write_plotly(fig: 'Figure', path: str):
    fig_json = fig.to_json()
    with open(path, 'w') as f:
        f.write(fig_json)


def write_matplotlib(fig: 'plt.Figure', path: str):
    fig.savefig(path)


def maybe_handle_plotly(obj, kwargs, show: ShowPlot = None):
    try:
        from plotly.graph_objs import Figure
        if isinstance(obj, Figure):
            fig = obj
            metadata = write_obj(fig, write=write_plotly, fmt='plotly', **kwargs)
            if show == 'png':
                img_bytes = fig.to_image(format='png')
                display(Image(img_bytes), metadata=metadata)
            elif show == 'html':
                html = fig.to_html()
                display(HTML(html), metadata=metadata)
            else:
                display(fig, metadata=metadata)
            return True
    except ImportError:
        pass
    return False


def maybe_handle_matplotlib(obj, kwargs):
    try:
        import matplotlib.pyplot as plt
        if isinstance(obj, plt.Figure):
            fig = obj
            metadata = write_obj(fig, write=write_matplotlib, fmt='png', **kwargs)
            display(fig, metadata=metadata)
            plt.close()
            return True
    except ImportError:
        pass
    return False


def export(
        obj,
        name: str = None,
        dir: str = None,
        git: Optional[GitConfig] = None,
        dvc: Optional[DvcConfig] = None,
        verbose: bool = False,
        show: ShowPlot = None,
        engine: ParquetEngine = None,
):
    if git is None:
        git = config.git
    if dvc is None:
        dvc = config.dvc
    if dir is None:
        dir = config.dir
    if show:
        if show not in ['html', 'png']:
            raise ValueError(f"Unrecognized `show` param: {show}")

    kwargs = dict(dir=dir, name=name, git=git, dvc=dvc, verbose=verbose)
    if isinstance(obj, pd.DataFrame):
        df = obj
        metadata = write_obj(obj=obj, write=write_parquet, fmt='parquet', **kwargs)
        html = df._repr_html_()
        return display(HTML(html), metadata=metadata)
    elif isinstance(obj, Image):
        metadata = write_obj(obj=obj, write=write_image, fmt='png', **kwargs)
        return display(obj, metadata=metadata)
    elif isinstance(obj, bytes):
        img = Image(obj)
        metadata = write_obj(obj=img, write=write_image, fmt='png', **kwargs)
        return display(img, metadata=metadata)
    elif not maybe_handle_plotly(obj, kwargs, show=show) and not maybe_handle_matplotlib(obj, kwargs):
        raise ValueError(f"Unrecognized export type: {type(obj)}")


def _ire_df_repr_html_(df):
    return export(df)


def push(
        git: bool = True,
        dvc: Optional[bool] = None,
        # Err on the side of letting the user know about Git/DVC pushes
        verbose: bool = True,
):
    if git:
        if has_unpushed_commits():
            run('git', 'push', verbose=verbose)
        else:
            if verbose:
                err("Git appears up to date, skipped push")
    if dvc or (dvc is None and config.dvc is not False):
        run('dvc', 'push', verbose=verbose)
